/*
 * AutoDAO - Generic DAO on steroids implementation for Java.
 *
 * Copyright 2008-2012  Marat Radchenko <slonopotamus@users.sourceforge.net>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.sf.autodao;

import java.io.Serializable;

import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.springframework.dao.DataAccessException;

/**
 * Typesafe CRUD interface.
 * <p/>
 * In order to use AutoDAO CRUD operations just inherit your DAO interface from this one.
 * <p/>
 * See <a href="http://autodao.sourceforge.net/guide.html#CRUD_operations">CRUD operations</a> for usage.
 * @param <E> entity type.
 * @param <K> primary key.
 * @since AutoDAO 0.1
 */
@SuppressWarnings({ "UnusedDeclaration" })
public interface Dao<E extends PersistentEntity<K>, K extends Serializable> {

  /**
   * Persist the new instance object into database. [create]
   * @param o transient object.
   * @throws IllegalArgumentException if given entity is already persistent.
   * @see org.hibernate.Session#persist(Object)
   */
  void create(@NotNull E o) throws DataAccessException, IllegalArgumentException;

  /**
   * Retrieve an object that was previously persisted to the database using the indicated id as primary key. [read]
   * @param id primary key
   * @return instance of object or <code>null</code> if object is not found.
   * @see org.hibernate.Session#get(Class, java.io.Serializable)
   */
  @Nullable
  E get(@NotNull K id) throws DataAccessException;

  /**
   * Update the given object. [update]
   * @param o instance of object to update.
   * @see org.hibernate.Session#update(Object)
   */
  void update(@NotNull E o) throws DataAccessException;

  /**
   * Remove an object from persistent storage in the database. [delete]
   * @param o object.
   * @see org.hibernate.Session#delete(Object)
   */
  void delete(@NotNull E o) throws DataAccessException;

  /**
   * Save or update object.
   * @param o object.
   * @see org.hibernate.Session#saveOrUpdate(Object)
   */
  void saveOrUpdate(@NotNull E o) throws DataAccessException;

  /**
   * Merge object with persistent object.
   * @param o object to merge.
   * @return merged object
   * @see org.hibernate.Session#merge(Object)
   */
  @NotNull
  E merge(@NotNull E o) throws DataAccessException;

  /**
   * Refresh the state of the instance from the database, overwriting changes made to the entity, if any.
   * @param o object.
   * @see org.hibernate.Session#refresh(Object)
   */
  void refresh(@NotNull final E o) throws DataAccessException;
}
