/*
 * AutoDAO - Generic DAO on steroids implementation for Java.
 *
 * Copyright 2008-2012  Marat Radchenko <slonopotamus@users.sourceforge.net>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.sf.autodao.impl;

import java.lang.reflect.Method;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import net.sf.autodao.Dao;
import net.sf.autodao.PersistentEntity;
import net.sf.autodao.QueryArgumentTransformer;
import org.aopalliance.aop.Advice;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.springframework.aop.framework.ProxyFactoryBean;
import org.springframework.beans.factory.BeanFactory;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.transaction.PlatformTransactionManager;
import org.springframework.transaction.TransactionDefinition;
import org.springframework.transaction.interceptor.DefaultTransactionAttribute;
import org.springframework.transaction.interceptor.TransactionAttribute;
import org.springframework.transaction.interceptor.TransactionAttributeSource;
import org.springframework.transaction.interceptor.TransactionInterceptor;
import org.springframework.util.Assert;

/**
 * Base ORM-agnostic DAO factory.
 * <p/>
 * NOT FOR PUBLIC USE.
 */
public abstract class AbstractDaoFactoryBean extends ProxyFactoryBean implements InitializingBean, TransactionAttributeSource {
  @NotNull
  private static final String QUERY_ARGUMENT_TRANSFORMERS_BEAN_NAME = "autodaoQueryArgumentTransformers";
  @NotNull
  private static final String LEGACY_QUERY_ARGUMENT_TRANSFORMER_BEAN_NAME = "autodaoQueryArgumentTransformer";
  private static final Log LOG = LogFactory.getLog(AbstractDaoFactoryBean.class);
  private static final TransactionAttribute MANDATORY_TRANSACTION_ATTRIBUTE = new DefaultTransactionAttribute(TransactionDefinition.PROPAGATION_MANDATORY);

  @NotNull
  static final Map<Class<?>, QueryArgumentTransformer> DEFAULT_TRANSFORMERS = defaultTransformers();

  @NotNull
  private static Map<Class<?>, QueryArgumentTransformer> defaultTransformers() {
    final Map<Class<?>, QueryArgumentTransformer> result = new HashMap<Class<?>, QueryArgumentTransformer>();

    ReadableInstantTransformer.register(result);

    return Collections.unmodifiableMap(result);
  }

  @NotNull
  private Map<Class<?>, QueryArgumentTransformer> transformers = DEFAULT_TRANSFORMERS;
  @NotNull
  private final Class<?> daoInterface;
  @NotNull
  private final PlatformTransactionManager transactionManager;
  @Nullable
  private final Class<? extends PersistentEntity<?>> entityType;

  protected AbstractDaoFactoryBean(@NotNull Class<?> daoInterface,
                                   @Nullable Class<? extends PersistentEntity<?>> entityType,
                                   @NotNull PlatformTransactionManager transactionManager) {
    this.daoInterface = daoInterface;
    this.entityType = entityType;
    this.transactionManager = transactionManager;
  }

  @NotNull
  protected abstract FinderExecutor createDao(@Nullable final Class<?> entityType);

  protected abstract void validate(@NotNull final Class<?> daoInterface, @NotNull Map<Class<?>, QueryArgumentTransformer> transformers);

  @Override
  @SuppressWarnings({ "unchecked" })
  public final void setBeanFactory(final BeanFactory beanFactory) {
    transformers = new HashMap<Class<?>, QueryArgumentTransformer>();
    transformers.putAll(DEFAULT_TRANSFORMERS);
    if (beanFactory.containsBean(QUERY_ARGUMENT_TRANSFORMERS_BEAN_NAME)) {
      transformers.putAll((Map<Class<?>, QueryArgumentTransformer>) beanFactory.getBean(QUERY_ARGUMENT_TRANSFORMERS_BEAN_NAME, Map.class));
    }
    if (beanFactory.containsBean(LEGACY_QUERY_ARGUMENT_TRANSFORMER_BEAN_NAME)) {
      LOG.warn(String.format("'%s' bean is no longer supported. Please, consult AutoDAO documentation.", LEGACY_QUERY_ARGUMENT_TRANSFORMER_BEAN_NAME));
    }
    super.setBeanFactory(beanFactory);
  }

  @Override
  public final void afterPropertiesSet() {
    validate(daoInterface, transformers);
    final Class<? extends PersistentEntity<?>> realEntityType = getEntityType(daoInterface, entityType);
    addInterface(daoInterface);
    setTarget(createDao(realEntityType));
    addAdvice(createTransactionInterceptor(transactionManager));
    addAdvice(new FinderIntroductionInterceptor(transformers));
  }

  @NotNull
  private Advice createTransactionInterceptor(@NotNull final PlatformTransactionManager transactionManager) {
    return new TransactionInterceptor(transactionManager, this);
  }

  @Nullable
  private static Class<? extends PersistentEntity<?>> getEntityType(@NotNull final Class<?> daoInterface,
                                                                    @Nullable final Class<? extends PersistentEntity<?>> explicit) {
    final Class<? extends PersistentEntity<?>> result;
    if (explicit == null && Dao.class.isAssignableFrom(daoInterface)) {
      result = autoDiscoverEntityType(daoInterface);
      Assert.notNull(result, "No entity type found for " + daoInterface);
    } else {
      result = explicit;
    }
    if (result != null) {
      Assert.isAssignable(PersistentEntity.class, result, "Invalid entity type: ");
    }
    return result;
  }

  @SuppressWarnings({ "unchecked" })
  private static Class<? extends PersistentEntity<?>> autoDiscoverEntityType(final Class<?> daoInterface) {
    //Automagically discover entity type from generics.
    for (final Type t : daoInterface.getGenericInterfaces()) {
      if (t instanceof ParameterizedType) {
        final ParameterizedType pt = (ParameterizedType) t;
        if (pt.getRawType() == Dao.class) {
          final Type type = pt.getActualTypeArguments()[0];
          if (type instanceof ParameterizedType) {
            //Special handling for cases like Dao<Entity<Foo>>.
            return (Class<? extends PersistentEntity<?>>) ((ParameterizedType) type).getRawType();
          } else {
            Assert.isInstanceOf(
                Class.class,
                type,
                "Don't know how to handle " + type + " entity type for " + daoInterface.getName() + ".\n"
                    + "Please update to latest AutoDAO release.\n"
                    + "If that doesn't help, please file a bugreport to AutoDAO bugtracker.");
            return (Class<? extends PersistentEntity<?>>) type;
          }
        }
      }
    }
    return null;
  }

  @Override
  public TransactionAttribute getTransactionAttribute(Method method, Class<?> targetClass) {
    return method.getDeclaringClass() == Object.class ? null : MANDATORY_TRANSACTION_ATTRIBUTE;
  }
}
