/*
 * AutoDAO - Generic DAO on steroids implementation for Java.
 *
 * Copyright 2008-2012  Marat Radchenko <slonopotamus@users.sourceforge.net>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.sf.autodao.impl;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import net.sf.autodao.Finder;
import net.sf.autodao.Named;
import net.sf.autodao.QueryArgumentTransformer;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.springframework.core.annotation.AnnotationUtils;
import static org.springframework.util.StringUtils.hasText;

/**
 * Performs finder method arguments validation.
 * <p/>
 * NOT FOR PUBLIC USE.
 */
public abstract class ParametersChecker implements Utils.ParameterCallback {
  private final boolean singleFind;
  @NotNull
  private final Method method;
  @NotNull
  private final Map<Class<?>, QueryArgumentTransformer> transformers;

  @NotNull
  private final Set<String> names = new HashSet<String>();
  private boolean limit;
  private boolean offset;
  private int indexed;

  protected ParametersChecker(@NotNull final Method method,
                              @NotNull final Map<Class<?>, QueryArgumentTransformer> transformers) {
    final Finder f = AnnotationUtils.getAnnotation(method, Finder.class);
    if (f == null)
      throw new SpecificationViolationException("Method is neither a finder method nor is implemented in AutoDAO", method);

    final int sum
        = (hasText(f.query()) ? 1 : 0)
        + (hasText(f.queryName()) ? 1 : 0)
        + (hasText(f.sqlQuery()) ? 1 : 0)
        + (hasText(f.sqlQueryName()) ? 1 : 0);
    if (sum < 1)
      throw new SpecificationViolationException("You need to specify at least one of: query, queryName, sqlQuery, sqlQueryName", method);
    else if (sum > 1)
      throw new SpecificationViolationException("You can specify only one of: query, queryName, sqlQuery, sqlQueryName", method);

    this.singleFind = validateReturnType(method);
    this.method = method;
    this.transformers = transformers;
  }

  @NotNull
  protected final Method getMethod() {
    return method;
  }

  @Nullable
  protected abstract Class<?> getExpectedType(int paramIndex);

  @Nullable
  protected abstract Class<?> getExpectedType(@NotNull String paramName);

  protected abstract int getExpectedOrdinalParametersCount();

  @Nullable
  protected abstract Set<String> getExpectedNamedParameters();

  /**
   * @param method method to validate.
   * @return <code>true</code> if method is single-find, <code>false</code> otherwise.
   * @throws SpecificationViolationException
   *
   */
  private static boolean validateReturnType(@NotNull final Method method) {
    final Finder f = AnnotationUtils.findAnnotation(method, Finder.class);
    final boolean singleFind = !Collection.class.isAssignableFrom(method.getReturnType());
    if (!singleFind) {
      final Class<?> returnType = method.getReturnType();
      if (!returnType.isAssignableFrom(f.returnAs())) {
        throw new SpecificationViolationException(
            "@Finder.returnAs value (" + f.returnAs().getName()
                + ") cannot be cast to method return type (" + returnType.getName() + ")", method
        );
      } else if (f.returnAs() != List.class) {
        if (f.returnAs().isInterface()) {
          throw new SpecificationViolationException(
              "@Finder.returnAs value (" + f.returnAs().getName() + ") cannot be an interface", method
          );
        } else {
          try {
            f.returnAs().getConstructor();
          } catch (NoSuchMethodException e) {
            throw new SpecificationViolationException(
                "@Finder.returnAs value (" + f.returnAs().getName() + ") doesn't have no-arg constructor.", method
            );
          }
        }
      }
    }
    return singleFind;
  }

  public void check() {
    Utils.visitMethodParameters(method, this);


    int queryIndexedCount = getExpectedOrdinalParametersCount();
    if (queryIndexedCount >= 0 && indexed != queryIndexedCount)
      throw new SpecificationViolationException(String.format("Not enough indexed params (query has %s but method has %s)", queryIndexedCount, indexed), method);

    final Set<?> expectedNamedParameters = getExpectedNamedParameters();
    if (expectedNamedParameters != null && !names.equals(expectedNamedParameters))
      throw new SpecificationViolationException(String.format("Wrong named parameters (query has %s but method has %s)", expectedNamedParameters, names), method);
  }

  @Override
  public void visit(final int index, @NotNull final Class<?> type, @NotNull final Annotation[] annotations) {
    boolean processed = false;
    if (Utils.getLimit(annotations) != null) {
      limitCheck(type);
      processed = true;
    }
    if (Utils.getOffset(annotations) != null) {
      offsetCheck(type);
      processed = true;
    }
    final Named named = Utils.getNamed(annotations);
    if (processed) {
      if (named != null) {
        throw new SpecificationViolationException("@Named cannot be present on @Limit/@Offset param", method);
      }
    } else {
      if (named == null) {
        if (!names.isEmpty()) {
          throw new SpecificationViolationException("You're not allowed to mix @Named and indexed params", method);
        }
        final int paramIndex = index + 1
            - (limit ? 1 : 0)
            - (offset ? 1 : 0);
        final Class<?> expectedType = getExpectedType(paramIndex);
        if (expectedType != null) {
          final Class<?> actualType = getActualParameterType(type, false);
          if (!expectedType.isAssignableFrom(actualType)) {
            final String msg = String.format("Indexed parameter '%s' of type %s cannot be assigned to query parameter of type %s", paramIndex, actualType.getName(), expectedType.getName());
            throw new SpecificationViolationException(msg, method);
          }
        }
        indexed++;
      } else {
        if (indexed > 0) {
          throw new SpecificationViolationException("You're not allowed to mix @Named and indexed params", method);
        }
        final String name = named.value();
        final Class<?> expectedType = getExpectedType(name);
        if (expectedType != null) {
          final Class<?> actualClass = getActualParameterType(type, true);
          if (!Collection.class.isAssignableFrom(actualClass) && !expectedType.isAssignableFrom(actualClass)) {
            final String msg = String.format("Named parameter '%s' of type %s cannot be assigned to query parameter of type %s", name, actualClass.getName(), expectedType.getName());
            throw new SpecificationViolationException(msg, method);
          }
        }
        if (!names.add(name)) {
          throw new SpecificationViolationException("Duplicate named parameters '" + name + "'", method);
        }
      }
    }
  }

  @NotNull
  private Class<?> getActualParameterType(@NotNull final Class<?> declaredType, final boolean handleArrays) {
    final QueryArgumentTransformer transformer = transformers.get(declaredType);
    final Class<?> transformedType = transformer == null ? declaredType : transformer.getTargetType();
    // TODO: check collection params too
    final Class<?> arrayUnwrappedType = handleArrays && transformedType.isArray() ? transformedType.getComponentType() : transformedType;
    return primitive(arrayUnwrappedType);
  }

  /**
   * Copied from commons-binutils (org.apache.commons.beanutils.converters.AbstractConverter).
   * <p/>
   * Change primitive Class types to the associated wrapper class.
   * @param type The class type to check.
   * @return The converted type.
   */
  @SuppressWarnings({ "IfStatementWithTooManyBranches", "ObjectEquality" })
  @NotNull
  private static Class<?> primitive(@NotNull final Class<?> type) {
    if (!type.isPrimitive()) {
      return type;
    } else if (type == Integer.TYPE) {
      return Integer.class;
    } else if (type == Double.TYPE) {
      return Double.class;
    } else if (type == Long.TYPE) {
      return Long.class;
    } else if (type == Boolean.TYPE) {
      return Boolean.class;
    } else if (type == Float.TYPE) {
      return Float.class;
    } else if (type == Short.TYPE) {
      return Short.class;
    } else if (type == Byte.TYPE) {
      return Byte.class;
    } else if (type == Character.TYPE) {
      return Character.class;
    } else {
      return type;
    }
  }

  private void limitCheck(final Class<?> type) {
    limitOffsetCheck(limit, singleFind, type, method);
    limit = true;
  }

  private void offsetCheck(final Class<?> type) {
    limitOffsetCheck(offset, singleFind, type, method);
    offset = true;
  }

  private void limitOffsetCheck(final boolean value,
                                final boolean singleFind,
                                final Class<?> type,
                                final Method m) {
    if (value) {
      throw new SpecificationViolationException("@Limit/@Offset can appear only on one parameter", m);
    } else if (singleFind) {
      throw new SpecificationViolationException("@Limit/@Offset isn't allowed on single-object finders", m);
    }
    if (type != Integer.class && type != Integer.TYPE) {
      throw new SpecificationViolationException(
          "@Limit and @Offset argument type must be either int or java.lang.Integer", m);
    }
  }
}
