/*
 * AutoDAO - Generic DAO on steroids implementation for Java.
 *
 * Copyright 2008-2012  Marat Radchenko <slonopotamus@users.sourceforge.net>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.sf.autodao.impl.hibernate

import java.lang.reflect.Method
import net.sf.autodao.QueryArgumentTransformer
import net.sf.autodao.impl.ParametersChecker
import net.sf.autodao.impl.SpecificationViolationException
import net.sf.autodao.impl.Utils
import org.hibernate.QueryParameterException
import org.hibernate.SessionFactory
import org.hibernate.type.Type
import org.jetbrains.annotations.NotNull
import org.jetbrains.annotations.Nullable

final class HibernateParametersChecker extends ParametersChecker {

  private final metadata;

  static void check(@NotNull SessionFactory sessionFactory, @NotNull Method method, @NotNull Map<Class<?>, QueryArgumentTransformer> transformers) {
    new HibernateParametersChecker(sessionFactory, method, transformers).check();
  }

  private HibernateParametersChecker(@NotNull SessionFactory sessionFactory, @NotNull Method method, @NotNull Map<Class<?>, QueryArgumentTransformer> transformers) {
    super(method, transformers);
    this.metadata = Utils.visitQuery(method, new ParameterMetadataExtractor(sessionFactory, method));
  }

  @Nullable
  @Override
  protected Class<?> getExpectedType(@NotNull String paramName) {
    def descriptor;
    try {
      descriptor = metadata.getNamedParameterDescriptor(paramName);
    } catch (final QueryParameterException e) {
      throw new SpecificationViolationException(String.format("Query doesn't have named parameter '%s'", paramName), getMethod(), e);
    }
    final Type expectedType = descriptor.getExpectedType();
    // The only case I know when this happens is SQL query.
    if (expectedType == null)
      return null;

    return expectedType.getReturnedClass();
  }

  @Override
  protected int getExpectedOrdinalParametersCount() {
    return metadata.getOrdinalParameterCount();
  }

  @Override
  protected Set<String> getExpectedNamedParameters() {
    return metadata.getNamedParameterNames();
  }

  @Nullable
  @Override
  protected Class<?> getExpectedType(int paramIndex) {
    try {
      return metadata.getOrdinalParameterDescriptor(paramIndex).getExpectedType().getReturnedClass();
    } catch (final QueryParameterException e) {
      throw new SpecificationViolationException("Query doesn't have enough indexed parameters", getMethod(), e);
    } catch (final IndexOutOfBoundsException e) {
      throw new SpecificationViolationException("Query doesn't have enough indexed parameters", getMethod(), e);
    }
  }
}

final class ParameterMetadataExtractor implements Utils.QueryVisitor {
  private final SessionFactory sessionFactory;
  private final Method method;

  ParameterMetadataExtractor(@NotNull SessionFactory sessionFactory, @NotNull final Method method) {
    this.sessionFactory = sessionFactory;
    this.method = method;
  }

  def visitNamedQuery(@NotNull String queryName) {
    def query = sessionFactory.getNamedQuery(queryName)
    if (query == null)
      throw new SpecificationViolationException("No named query '" + queryName + "' found.", method);
    return sessionFactory.getQueryPlanCache().getHQLQueryPlan(query.getQuery(), false, Collections.emptyMap()).getParameterMetadata();
  }

  def visitQuery(@NotNull String query) {
    try {
      return sessionFactory.getQueryPlanCache().getHQLQueryPlan(query, false, Collections.emptyMap()).getParameterMetadata();
    } catch (final Exception e) {
      throw new SpecificationViolationException("Invalid query: " + query, method, e);
    }
  }

  def visitSQLNamedQuery(@NotNull String queryName) {
    def query = sessionFactory.getNamedSQLQuery(queryName)
    if (query == null)
      throw new SpecificationViolationException("No named SQL query '" + queryName + "' found.", method);
    return sessionFactory.getQueryPlanCache().getSQLParameterMetadata(query.getQuery());
  }

  def visitSQLQuery(@NotNull String query) {
    return sessionFactory.getQueryPlanCache().getSQLParameterMetadata(query);
  }
}

